// -------------------------------------------------------------------------
//
//                                    R E S P
//                                    =======
//
// (peter Rosendahls Erstes Schach Programm).
//
//
//  Datei              : console.cpp
//                       Konsolenclient fr RESP.
//
//  Anfang des Projekts: So, 8.Juli, 2001
//  Sprache            : C++
//  Verwendete Compiler: Visual C++ 6.0 SP5 / gcc 3.01
//  Betriebssystem(e)  : Win2000 SP2 / Linux 2.4.4
//  Version            : 0.19
// -------------------------------------------------------------------------
//
// resp (peter _r_osendahls _e_rstes _s_chach_p_rogramm) 
// $Id: console.cpp,v 1.30 2002/06/11 21:00:49 Peter Rosendahl Exp $
// Copyright (C) 2001 Peter Rosendahl
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// Email: p_rosendahl@t-online.de
// -------------------------------------------------------------------------

#include "basic_stuff.h"
#include "controller.h"
#include "xboard.h"
#include <iostream>
#include <limits>

//#define XBOARD_INTERFACE_TEST

#ifndef LINUX
#pragma warning (disable:4786) // Visual C++ Faxen
#endif

using namespace std;

void title()
{
    //cout << endl << endl << endl;    
    out << "#    //" << endl;
    out << "#    ////  ////   ///  ////" << endl;
    out << "#   //   ////    /    / //"  << endl;
    out << "#  //    ///// ///   //"  << endl;
    out << "#" << endl;
    out << "# resp  ... peter (r)osendahls (e)rstes (s)chach (p)rogramm" << endl;
    out << "# " << endl;
    out << "# VERSION: " << VERSION_STRING << endl;
    out << "# (c) p.rosendahl, 2001-2002" << endl;
    out << "#" << endl;
    out << "#" << endl;

}


void dumpStatus(int status)
{
    if (status & ST_GAME_OVER)
        out << "GAME OVER  ";

    if (status & ST_CANCELED)
    {
        out << "GAME CANCELED" << endl;
        return;
    }

    if (status & ST_RESIGN)
    {
        if (status & ST_WHITE_TO_MOVE)
            out << "BLACK RESIGNS" << endl;
        else
            out << "WHITE RESIGNS" << endl;

        return;
    }
    if (status & ST_REMIS)
        out << "REMIS  ";
    else
        if (status & ST_WHITE_WON)
            out << "WHITE WON  ";
        else
            out << "BLACK WON  ";

    out <<  endl;
}


#ifdef LINUX
int main(int argc, char* argv[])
#else 
int __cdecl main(int argc, char* argv[]) // Warnung wg. __fastcall
#endif
{

    // Logging setzen
    out.setLogging(GlobalGameOptions::getHandle()->loggingEnabled);

    // Eventuell Win(X)Board-Engine?
    #ifndef XBOARD_INTERFACE_TEST
    if (argc > 1)
        if (strcmp(argv[1],"-xboard") == 0)
    #endif
        {
            // XBoard Interface starten;
            GlobalGameOptions::getHandle()->outStyle = GlobalGameOptions::XBOARD;
            XBoardInterface* pxb = XBoardInterface::getHandle();
            pxb->run();

            return 0;
        }

    Controller& control = Controller::getHandle();
    title();
    control.command("new");
    

    while (true) {
        out <<  "RESP> ";


        char s[256];
        cin.getline(s,80);

        if (out.getLogging())
            out.logfile() << s << endl;

        // ----------------------------------------------------------
        // Spezialfall XBoard sendet den Befehl 'xboard':
        // In diesem Fall wird das Xboard Interface gestartet.

        string tmp(s);
        if (tmp == "xboard")
        {
            // Ausgaben ab jetzt im XBoard Format
            GlobalGameOptions::getHandle()->outStyle = GlobalGameOptions::XBOARD;
	        cout << endl;

            if (out.getLogging())
            {
                out.logfile() << "#" << endl;
                out.logfile() << "# Switching to Winboard/XBoard Mode..." << endl;
                out.logfile() << "#" << endl;
            }
            XBoardInterface* pxb = XBoardInterface::getHandle();
            pxb->run();
            return 0;
        }
        // ----------------------------------------------------------

        
        try {
            control.command(s);
        }
        catch (BadCmdException e) {
            out << "Error - " << e.what() << endl;
        }

        if (control.getStatus() & ST_GAME_OVER)
            dumpStatus(control.getStatus());

        
        if (control.getStatus() & ST_CANCELED)
            break;

        if (control.getStatus() & ST_CHECK)
            out << "CHECK!" << endl;
    }

    //dumpStatus(control.getStatus());

    out << "Bye." << endl;


    return 0;
}
