// -------------------------------------------------------------------------------------
//
//                                    R E S P
//                                    =======
//
// (peter Rosendahls Erstes Schach Programm).
//
//
//  Datei              : eval.cpp
//                       Evaluierungsklasse Eval; zweite,vollkommen berarbeitete 
//                       Version.
//                       Im Gegensatz zur Vorgngerversion sind die Features keine
//                       Klassen mehr, wie z.Bsp. class PawnMaterial : public Feature,
//                       sondern komplett in der Evaluierungsfunktion integrierte
//                       Verfahren.
//                       Um dennoch flexible Parameter fr Lernalgorithmen zu haben,
//                       werden die Feature-Gewichte in einem Array abgespeichert.
//                       
//
//  Anfang             : Freitag 2.November, 2001
//  Sprache            : C++
//  Verwendete Compiler: Visual C++ 6.0 SP5 / gcc 3.01
//  Betriebssystem(e)  : Win2000 SP2 / Linux 2.4.4
//  Version            : 0.19
// -------------------------------------------------------------------------------------
//
// resp (peter _r_osendahls _e_rstes _s_chach_p_rogramm) 
// $Id: eval.cpp,v 1.65 2002/06/11 19:04:44 Peter Rosendahl Exp $
// Copyright (C) 2001 Peter Rosendahl
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// Email: p_rosendahl@t-online.de
// -------------------------------------------------------------------------

#include "eval.h"
#include "phash.h"
#include "basic_stuff.h"
#include "bitboard.h"
#include "recognizer.h"
#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>
using namespace std;



// ----------------------------------------------------------------
//  S t a t i s c h e  E v a l u i e ru n g s t a b e l l e n
// ----------------------------------------------------------------

//  ----------------
//    B a u e r n
//  ----------------


int Eval::posval_pawn_op[64] =
{
     0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0,
    -4, -4,  2,  6,  6,-10,-10,-10,
    -3, -3,  2,  8, 10, -9, -9, -9,
    -2, -2,  1, 15, 16, -6, -6, -6,
    -1, -1,  1,  6,  6, -3, -3, -3,
     0,  0,  0,-10,-10,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0
};



int Eval::posval_pawn_emid[64] =
{
     0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  4,  4,  0,  0,  0,
     0,  0,  2,  8,  8,  0,  0,  0,
     5,  5,  7, 12, 12,  4,  4,  0,
     3,  2,  3,  6,  6, -2, -1, -1,
     0,  0,  0,-20,-20,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0
};

int Eval::posval_pawn_lmid[64] =
{
     0,  0,  0,  0,  0,  0,  0,  0,
     9, 10, 10, 10, 10, 10, 10,  9,
     7,  8,  8,  8,  8,  8,  8,  7,
     5,  6,  6,  6,  6,  6,  6,  5,
     3,  4,  4,  4,  4,  4,  4,  3,
     1,  2,  2,  2,  2,  2,  2,  1,
     0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0
};

int Eval::posval_pawn_end[64] =
{
     0,  0,  0,  0,  0,  0,  0,  0,
    10, 11, 11, 11, 11, 11, 11, 10,
     8,  9,  9,  9,  9,  9,  9,  8,
     6,  7,  7,  7,  7,  7,  7,  6,
     4,  4,  4,  4,  4,  4,  4,  4,
     2,  2,  2,  2,  2,  2,  2,  2,
     0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0
};

//  ----------------
//   K  n i g 
//  ----------------
int Eval::posval_king_op[64] = {
   -99,-99,-99,-99,-99,-99,-99,-99,
   -99,-99,-99,-99,-99,-99,-99,-99,
   -99,-99,-99,-99,-99,-99,-99,-99,
   -99,-99,-99,-99,-99,-99,-99,-99,
   -99,-99,-99,-99,-99,-99,-99,-99,
   -80,-80,-80,-80,-80,-80,-80,-80,
   -35,-35,-35,-45,-45,-35,-35,-35,
    10, 10, 10,-15,  0,-15, 14, 12
};
int Eval::posval_king_emid[64] = {
   -30,-30,-30,-30,-30,-30,-30,-30,
   -30,-30,-30,-30,-30,-30,-30,-30,
   -30,-30,-30,-30,-30,-30,-30,-30,
   -30,-30,-30,-30,-30,-30,-30,-30,
   -30,-30,-30,-30,-30,-30,-30,-30,
   -30,-30,-30,-30,-30,-30,-30,-30,
   -15,-15,-15,-15,-15,-15,-15,-15,
    10, 10, 10,-15,  0,-15, 12, 12
};
int Eval::posval_king_lmid[64] = {
     0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0,
    -1, -1, -1,  0,  0, -1, -1, -1,
     1,  5,  1,  0,  0,  1,  5,  1
};
int Eval::posval_king_end[64] = {
     0,  6, 12, 18, 18, 12,  6,  0,
     6, 12, 18, 24, 24, 18, 12,  6,
    12, 18, 24, 30, 30, 24, 18, 12,
    18, 24, 30, 36, 36, 30, 24, 18,
    18, 24, 30, 36, 36, 30, 24, 18,
    12, 18, 24, 30, 30, 24, 18, 12,
     6, 12, 18, 24, 24, 18, 12,  6,
     0,  6, 12, 18, 18, 12,  6,  0
};

//  -----------------
//   S p r i n g e r
//  -----------------
int Eval::posval_knight_op_emid[64] = {
    -8, -5, -5, -5, -5, -5, -5, -8,
    -5,  0,  0,  3,  3,  0,  0, -5,
    -5,  0,  5,  5,  5,  5,  0, -5,
    -5,  0,  5,  9,  9,  5,  0, -5,
    -5,  0,  5,  9,  9,  5,  0, -5,
    -5,  0,  5,  5,  5,  5,  0, -5,
    -5,  0,  0,  3,  3,  0,  0, -5,
    -7, -8, -4, -4, -4, -4, -8, -7
};

int Eval::posval_knight_lmid_end[64] = {
   - 5,  0,  0,  0,  0,  0,  0, -5,
     0,  0,  0,  1,  1,  0,  0,  0,
     0,  0,  1,  3,  3,  1,  0,  0,
     0,  1,  3,  5,  5,  3,  1,  0,
     0,  1,  3,  5,  5,  3,  1,  0,
     0,  0,  1,  3,  3,  1,  0,  0,
     0,  0,  0,  1,  1,  0,  0,  0,
   - 5,  0,  0,  0,  0,  0,  0, -5
};

//  -----------------
//    L  u f e r 
//  -----------------
int Eval::posval_bishop_op_emid[64] = {
    -5, -5, -5, -5, -5, -5, -5, -5,
    -5,  7,  5,  7,  7,  5,  7, -5,
    -5,  5,  3,  6,  6,  3,  5, -5,
    -5,  3,  6,  9,  9,  6,  3, -5,
    -5,  3,  6,  9,  9,  6,  3, -5,
    -5,  5,  3,  4,  4,  3,  5, -5,
    -5,  7,  5,  5,  5,  5,  7, -5,
    -5, -8, -5, -5, -5, -5, -8, -5,
};

int Eval::posval_bishop_lmid_end[64] = {
     0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  4,  4,  0,  0,  0,
     0,  0,  4,  8,  8,  4,  0,  0,
     0,  4,  8, 12, 12,  8,  4,  0,
     0,  4,  8, 12, 12,  8,  4,  0,
     0,  0,  4,  8,  8,  4,  0,  0,
     0,  0,  0,  4,  4,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0,
};

//  -----------------
//      T  r m e
//  -----------------
int Eval::posval_rook_op[64] = {
     0,  0,  0,  0,  0,  0,  0,  0,
    34, 34, 34, 34, 34, 34, 34, 34,
     0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  5,  5,  3,  0,  0
};
int Eval::posval_rook_emid[64] = {
     0,  0,  0,  0,  0,  0,  0,  0,
    28, 28, 28, 28, 28, 28, 28, 28,
     0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  5,  5,  3,  0,  0
};
int Eval::posval_rook_lmid[64] = {
     0,  0,  0,  0,  0,  0,  0,  0,
    24, 24, 24, 24, 24, 24, 24, 24,
     0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0
};
int Eval::posval_rook_end[64] = {
     0,  0,  0,  0,  0,  0,  0,  0,
    15, 15, 15, 15, 15, 15, 15, 15,
     0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0,
     0,  0,  0,  0,  0,  0,  0,  0
};


// --------------------------------------------------------------
//  Score-Tabellen
// --------------------------------------------------------------

int Eval::score_passed_pawn[8] = {
    0,8,14,30,60,100,130,0
};

int Eval::score_block_passer[8] = {
    0,2,4,6,12,25,32,0
};

int Eval::score_out_ppawns[8] = {
    1,2,6,7,11,13,15,17
};


// --------------------------------------------------------------
//  Definition Eval Object
// --------------------------------------------------------------
Eval Eval::evalObject;


// --------------------------------------------------------------
//  Konstruktor
// --------------------------------------------------------------

Eval::Eval()
{
    cout << "#Initializing Eval" << endl;
    initPieceVal();
    cout << "#Eval ready" << endl;
}


// --------------------------------------------------------------
// Initialisierung der Feature Gewichte
// --------------------------------------------------------------


void Eval::initPieceVal()
{
    // piece_val aktualisieren
    piece_val[PAWN]     = PAWN_MATERIAL;
    piece_val[BISHOP]   = BISHOP_MATERIAL;
    piece_val[KNIGHT]   = KNIGHT_MATERIAL;
    piece_val[ROOK]     = ROOK_MATERIAL;
    piece_val[QUEEN]    = QUEEN_MATERIAL;
}

// --------------------------------------------------------------
//  Bestimmen der aktuellen Spielphase (Erffnung, Mittelspiel,
//  Endspiel)
// --------------------------------------------------------------
inline int Eval::getStage(Board *pB)
{
    int n_pawns = pB->stats.nBlackPieces[PAWN] + 
                  pB->stats.nWhitePieces[PAWN];
    int n_pieces = pB->stats.nBlackPiecesTotal + 
                   pB->stats.nWhitePiecesTotal - n_pawns;


    if (n_pawns > 12 && n_pieces > 12)
        return OPENING;
    else if (n_pieces >= 12 || (n_pawns > 10 && n_pieces > 10) )
        return EARLY_MIDGAME;
    else if (n_pawns > 6 && n_pieces > 6)
        return LATE_MIDGAME;
    else
        return ENDGAME;
}


// --------------------------------------------------------------
//                         e v a l
// --------------------------------------------------------------
//  Evaluierungsfunktions fr resp. Neben der aktuellen Position
//  aus pB wird das alpha-beta Fenster der Suche bergeben. Dies
//  ermglicht 'eval' eine "Lazy Evaluation", falls in einem
//  frhen Stadium der Evaluierung klar ist, dass 
//
//      score <= alpha     ODER   score >= beta
//  
//  gilt.
// --------------------------------------------------------------
int Eval::eval(Board* pB, int alpha, int beta)
{
    // Aktuelle Spielphase bestimmen
    int stage = getStage(pB);
    switch (stage)
    {
    case OPENING:

#ifdef DEBUG_EVAL
        cout << ".................OPENING EVALUATION" << endl;
#endif
        return eval_op(pB,alpha,beta);
    case EARLY_MIDGAME:

#ifdef DEBUG_EVAL
        cout << ".................EARLY MIDGAME EVALUATION" << endl;
#endif
        return eval_emid(pB,alpha,beta);
    case LATE_MIDGAME:

#ifdef DEBUG_EVAL
        cout << ".................LATE MIDGAME EVALUATION" << endl;
#endif
        return eval_lmid(pB,alpha,beta);
    default:

#ifdef DEBUG_EVAL
        cout << ".................ENDGAME EVALUATION" << endl;
#endif
        return eval_end(pB,alpha,beta);
    }
}

// --------------------------------------------------------------
//  Evaluierungsfunktion nur Material
// --------------------------------------------------------------

int Eval::getMatScore(Board *pB)
{
    int score = 0;
    // --------------------------
    // Materialscore berechnen
    // --------------------------
    score = (pB->stats.nWhitePieces[PAWN] - pB->stats.nBlackPieces[PAWN]) *
                PAWN_MATERIAL;
    score += (pB->stats.nWhitePieces[BISHOP] - pB->stats.nBlackPieces[BISHOP]) *
                BISHOP_MATERIAL;
    score += (pB->stats.nWhitePieces[KNIGHT] - pB->stats.nBlackPieces[KNIGHT]) *
                KNIGHT_MATERIAL;
    score += (pB->stats.nWhitePieces[ROOK] - pB->stats.nBlackPieces[ROOK]) *
                ROOK_MATERIAL;
    score += (pB->stats.nWhitePieces[QUEEN] - pB->stats.nBlackPieces[QUEEN]) * 
                QUEEN_MATERIAL;


    if (pB->sideToMove() == WHITE)
        return score;
    else 
        return -score;

}

//  ---------------------------------------------------------
//       E v a l u i e r u n g s f u n k t i o n  f  r  
//       B a u e r n s t r u k t ur 
//  ---------------------------------------------------------

void Eval::pawn_eval(const Board *pB, PTTe*& p_pawn_info,int stage, 
                    const int posval_pawn[])
{
    static int whiteLAdvPawn[10];
    static int nWhitePawns[10];
    static int blackLAdvPawn[10];
    static int nBlackPawns[10];
    static bool punishedDblIso[8];

    PTTe pawn_info = PTTe(pB->get_zp2_hash(),0,stage,0,0);

    Bitboard set;
    int pos;

    int i;

#ifdef DEBUG_EVALPAWNS
    cout << "-PAWN EVAL --------------------------------------" << endl;
#endif


    // Initialisieren
    for (i = 0; i < 10; i++)
    {
        whiteLAdvPawn[i] = 0;
        nWhitePawns[i]   = 0;
        blackLAdvPawn[i] = 7;
        nBlackPawns[i]   = 0;
    }

    // -----------------------------
    //    D u r c h g a n g   ( I )
    // -----------------------------

    //    S c a n n e n   d e r  w e i s s e n  B a u e r n

    set = pB->whitePawns;

    int pscore = 0; // Position
    int sscore = 0; // Struktur
    while (set)
    {
        pos = set.msb();
        set ^= BBTables::bbmask[pos];

        pscore += posval_pawn[pos];

        int c = COL(pos); 
        int r = ROW(pos);

        if (r > whiteLAdvPawn[ c + 1 ])
            whiteLAdvPawn[ c + 1 ] = r;

        nWhitePawns[c + 1]++;

        if (nWhitePawns[c + 1] >= 2)
        {
            // Doppelbauern gefunden
            sscore -= DOUBLE_PAWN_PENALTY;

#ifdef DEBUG_EVALPAWNS
            cout << "WHITE DOUBLE PAWN ON COL = " << c ;
            cout << "  [ " << -DOUBLE_PAWN_PENALTY << " ]" << endl;
#endif
        }
    }

    // Halboffene Linien fr Schwarz nach pinfo speichern
    for (i=1; i <= 8; i++)
        if (!nWhitePawns[i])
            pawn_info.setHalfOpenWhite(i - 1);


    //    S c a n n e n  d e r  s c h w a r z e n  B a u e r n
    
    set = pB->blackPawns;

    while (set)
    {
        pos = set.lsb();
        set ^= BBTables::bbmask[pos];

        pscore -= posval_pawn[mirror(pos)];


        int c = COL(pos); 
        int r = ROW(pos);

        if (r < blackLAdvPawn[ c + 1 ])
            blackLAdvPawn[ c + 1 ] = r;

        nBlackPawns[c + 1]++;

        if (nBlackPawns[c + 1] >= 2)
        {
            // Doppelbauern gefunden
            sscore += DOUBLE_PAWN_PENALTY;

#ifdef DEBUG_EVALPAWNS
            cout << "BLACK DOUBLE PAWN ON COL = " << c ;
            cout << "  [ " << DOUBLE_PAWN_PENALTY << " ]" << endl;
#endif
        }
    }

    // Halboffene Linien fr Weiss nach pinfo speichern
    for (i=1; i <= 8; i++)
        if (!nBlackPawns[i])
            pawn_info.setHalfOpenBlack(i - 1);


    // ------------------------------
    //  D u r c h g a n g   ( I I )
    // ------------------------------
    
    // Strken/Schwchen der Bauernstruktur feststellen


    // *****************************************
    // *******  w e i s s e  B a u e r n *******
    // *****************************************

    // Flaggenfeld fr isolierte Doppelbauern
    for (i=0; i < 8; i++)
        punishedDblIso[i] = false;

    set = pB->whitePawns;
    while (set)
    {
        pos = set.msb();
        set ^= BBTables::bbmask[pos];

        // --------------------------------------
        //          Bauern-Duo
        // --------------------------------------
        // Dies sind einfach benachbarte Bauern
        // gleicher Farbe

        int c = COL(pos);
        int r = ROW(pos);

        if (c < 7)
        {
            if (BBTables::bbmask[pos+1] & pB->whitePawns)
            {
                sscore += PAWN_DUO_BONUS;

#ifdef DEBUG_EVALPAWNS
                cout << "WHITE DUO ON " << c << ", " << c+1;
                cout << "  [ " << PAWN_DUO_BONUS << " ]" << endl;
#endif
            }
        }


        bool left_partner = (c>0) ?  !pawn_info.isHalfOpenWhite(c-1) : 
                                      false;
                                      
        bool right_partner = (c<7) ? !pawn_info.isHalfOpenWhite(c+1) : 
                                      false;

        // ----------------------------------------
        //              Freibauern
        // ----------------------------------------
        // Bauern, die nicht mehr von gegnerischen
        // Bauern geschlagen werden knnen

        bool passed_pawn = false;
        if ((BBTables::w_ppawn_mask[pos] & pB->blackPawns) == 0)
        {
            passed_pawn = true;
            sscore += stage*score_passed_pawn[7-r]/4;
            pawn_info.setWPassed(c);

#ifdef DEBUG_EVALPAWNS
            cout << "PASSED WHITE PAWN: " << pos;
            cout << "  [ " << stage*score_passed_pawn[7-r]/4
                 << " ]" << endl;
#endif
        }

        // --------------------------------------
        //      Isolierter Bauer
        // --------------------------------------
        // Keine Bauern gleicher Farbe auf den
        // Nachbarlinien
        if (!(left_partner || right_partner))
        {
            // Nicht doppelt?
            if (nWhitePawns[c + 1] == 1)
            {
                sscore -= ISOLATED_PAWN_PENALTY;

#ifdef DEBUG_EVALPAWNS
                cout << "ISOLATED WHITE PAWN: " << pos;
                cout << "  [ " << -ISOLATED_PAWN_PENALTY << " ]" << endl;
#endif
            }
            else // Doppelt
            {
                if (!punishedDblIso[c])
                {
                    punishedDblIso[c] = true;
                    // Besonders bel sind isolierte Doppelbauern: 
                    // Als Strafe gibt es: DOUBLE_PAWN_PENALTY  und dazu 
                    // ISOLATED_DBLPAWN_PENALTY
                    sscore -= ISOLATED_DBLPAWN_PENALTY;

#ifdef DEBUG_EVALPAWNS
                    cout << "ISOLATED WHITE DBLPAWN: " << pos;
                    cout << "  [ " << -ISOLATED_DBLPAWN_PENALTY << " ]" << endl;
#endif
                }
            }
            
        }

        // --------------------------------------
        //       Rckstndige Bauern
        // --------------------------------------
        // Bauern ohne Schutz von freundlichen
        // Bauern; Freibauern werden ausgeschlossen
        else if (!passed_pawn && (!left_partner || whiteLAdvPawn[c] < r)  && 
                 (!right_partner || whiteLAdvPawn[c+2] < r))
        {
            sscore -= BACKWARD_PAWN_PENALTY;

#ifdef DEBUG_EVALPAWNS
            cout << "BACKWARD WHITE PAWN: " << pos;
            cout << "  [ " << -BACKWARD_PAWN_PENALTY << " ]" << endl;
#endif
        }
    
    
    }

    // *****************************************
    // ****** s c h w a r z e  B a u e r n *****
    // *****************************************

    // Flaggenfeld fr isolierte Doppelbauern
    for (i=0; i < 8; i++)
        punishedDblIso[i] = false;

    set = pB->blackPawns;
    while (set)
    {
        pos = set.lsb();
        set ^= BBTables::bbmask[pos];

        // --------------------------------------
        //          Bauern-Duo
        // --------------------------------------
        // Dies sind einfach benachbarte Bauern
        // gleicher Farbe

        int c = COL(pos);
        int r = ROW(pos);

        if (c < 7)
        {
            if (BBTables::bbmask[pos+1] & pB->blackPawns)
            {
                sscore -= PAWN_DUO_BONUS;

#ifdef DEBUG_EVALPAWNS
                cout << "BLACK DUO ON " << c << ", " << c+1;
                cout << "  [ " << -PAWN_DUO_BONUS << " ]" << endl;
#endif
            }
        }


        bool left_partner = (c>0) ?  !pawn_info.isHalfOpenBlack(c-1) : 
                                      false;
                                      
        bool right_partner = (c<7) ? !pawn_info.isHalfOpenBlack(c+1) : 
                                      false;

        // ----------------------------------------
        //              Freibauern
        // ----------------------------------------
        // Bauern, die nicht mehr von gegnerischen
        // Bauern geschlagen werden knnen

        bool passed_pawn = false;
        if ((BBTables::b_ppawn_mask[pos] & pB->whitePawns) == 0)
        {
            passed_pawn = true;
            sscore -= stage*score_passed_pawn[r]/4;
            pawn_info.setBPassed(c);

#ifdef DEBUG_EVALPAWNS
            cout << "PASSED BLACK PAWN: " << pos;
            cout << "  [ " << -stage*score_passed_pawn[r]/4
                 << " ]" << endl;
#endif
        }


        // --------------------------------------
        //      Isolierter Bauer
        // --------------------------------------
        // Keine Bauern gleicher Farbe auf den
        // Nachbarlinien
        if (!(left_partner || right_partner))
        {
            // Nicht doppelt?
            if (nBlackPawns[c + 1] == 1)
            {
                sscore += ISOLATED_PAWN_PENALTY;

#ifdef DEBUG_EVALPAWNS
                cout << "ISOLATED BLACK PAWN: " << pos;
                cout << "  [ " << ISOLATED_PAWN_PENALTY << " ]" << endl;
#endif
            }
            else // Doppelt
            {
                if (!punishedDblIso[c])
                {
                    punishedDblIso[c] = true;
                    // Besonders bel sind isolierte Doppelbauern: 
                    // Als Strafe gibt es: DOUBLE_PAWN_PENALTY  und dazu 
                    // ISOLATED_DBLPAWN_PENALTY
                    sscore += ISOLATED_DBLPAWN_PENALTY;

#ifdef DEBUG_EVALPAWNS
                    cout << "ISOLATED BLACK DBLPAWN: " << pos;
                    cout << "  [ " << +ISOLATED_DBLPAWN_PENALTY << " ]" << endl;
#endif
                }
            }
            
        }
        // --------------------------------------
        //       Rckstndige Bauern
        // --------------------------------------
        // Bauern ohne Schutz von freundlichen
        // Bauern; Freibauern werden ausgeschlossen
        else if (!passed_pawn && (!left_partner || blackLAdvPawn[c] > r)  && 
                 (!right_partner || blackLAdvPawn[c+2] > r))
        {
            sscore += BACKWARD_PAWN_PENALTY;

#ifdef DEBUG_EVALPAWNS
            cout << "BACKWARD BLACK PAWN: " << pos;
            cout << "  [ " << +BACKWARD_PAWN_PENALTY << " ]" << endl;
#endif
        }
    
    
    }

    // ----------------------------------------------------------------------
    //  D u r ch g a n g   (I I I)
    // ----------------------------------------------------------------------
    // Connected passed pawns

    // ----------
    //  Weiss
    // ----------
    int w_passed = pawn_info.w_passed;
    while (w_passed)
    {
        int c = BBTables::lsb16[w_passed];
        if (c >= 7) break;
        w_passed ^= 1<<c;

        if (w_passed & 1<<(c+1))    // Zwei Freibauern nebeneinander
        {
            Bitboard temp = pB->whitePawns & BBTables::filemask[c];
            if (!temp) break;
            int pos1 = temp.lsb();
            if (ROW(pos1) >3) continue;

            temp = pB->whitePawns & BBTables::filemask[c+1];
            if (!temp) break;
            int pos2 = temp.lsb();
            if (ROW(pos2) > 3) continue;

            if (abs(ROW(pos1)-ROW(pos2)) <= 1)
            {
#ifdef DEBUG_EVALPAWNS
                cout << "CONNECTED WHITE PASSED PAWNS AT " << pos1 << "," << pos2 << endl;
#endif
                sscore += CONN_PASSED_PAWNS*stage;
            }

        }

    }

    // ----------
    //  Schwarz
    // ----------
    int b_passed = pawn_info.b_passed;
    while (b_passed)
    {
        int c = BBTables::lsb16[b_passed];
        if (c >= 7) break;
        b_passed ^= 1<<c;

        if (b_passed & 1<<(c+1))    // Zwei Freibauern nebeneinander
        {
            Bitboard temp = pB->blackPawns & BBTables::filemask[c];
            if (!temp) break;
            int pos1 = temp.msb();
            if (ROW(pos1) < 4) continue;

            temp = pB->blackPawns & BBTables::filemask[c+1];
            if (!temp) break;
            int pos2 = temp.msb();
            if (ROW(pos2) < 4) continue;

            if (abs(ROW(pos1)-ROW(pos2)) <= 1)
            {
#ifdef DEBUG_EVALPAWNS
                cout << "CONNECTED BLACK PASSED PAWNS AT " << pos1 << "," << pos2 << endl;
#endif
                sscore -= CONN_PASSED_PAWNS*stage;
            }

        }

    }

    // ----------------------------------------------------------------------
    //  Outside passed pawns
    // ----------------------------------------------------------------------

    // nur late midgame oder endgame + Bauern muessen vorhanden sein
    if (stage > 1 && (pB->stats.matBlack & 1) && (pB->stats.matWhite&1)) 
    {
        int w_out_pp = 0;
        int b_out_pp = 0;

        // -----------------
        //      Weiss
        // -----------------
        int w_passed = pawn_info.w_passed;
        int b_pawns  = pawn_info.getBPawns() & ~pawn_info.getBPassed();

        if (b_pawns && w_passed)
        {
            int col_mostright_bpawn = BBTables::msb16[b_pawns];

            if (col_mostright_bpawn <= 5)
            {
                int col_mostright_wpasser = BBTables::msb16[w_passed];
                if (col_mostright_wpasser - col_mostright_bpawn >= 2)
                {
                    w_out_pp = col_mostright_wpasser - col_mostright_bpawn;
                }
            }

            int col_mostleft_bpawn = BBTables::lsb16[b_pawns];
            if (col_mostleft_bpawn >= 2)
            {
                int col_mostleft_wpasser = BBTables::lsb16[w_passed];
                if (col_mostleft_bpawn - col_mostleft_wpasser >= 2   && 
                    col_mostleft_bpawn - col_mostleft_wpasser >= w_out_pp)
                {
                    w_out_pp = col_mostleft_bpawn - col_mostleft_wpasser;
                }
            }
        }
        // -----------------
        //      Schwarz
        // -----------------
        int b_passed = pawn_info.b_passed;
        int w_pawns  = pawn_info.getWPawns() & ~pawn_info.getWPassed();
        int col_mostright_wpawn = BBTables::msb16[w_pawns];

        if (w_pawns && b_passed)
        {
            if (col_mostright_wpawn <= 5)
            {
                int col_mostright_bpasser = BBTables::msb16[b_passed];
                if (col_mostright_bpasser >= col_mostright_wpawn + 2)
                {
                    b_out_pp = col_mostright_bpasser - col_mostright_wpawn;
                }
            }

            int col_mostleft_wpawn = BBTables::lsb16[w_pawns];
            if (col_mostleft_wpawn >= 2)
            {
                int col_mostleft_bpasser = BBTables::lsb16[b_passed];
                if (col_mostleft_wpawn - col_mostleft_bpasser >= 2 && 
                    col_mostleft_wpawn - col_mostleft_bpasser >= b_out_pp)
                {
                    b_out_pp = col_mostleft_wpawn - col_mostleft_bpasser;
                }
            }
        }
#ifdef DEBUG_EVALPAWNS
        cout << "OUTSIDEPASSED PAWNS - WHITE DIST.= " << w_out_pp << endl;
        cout << "OUTSIDEPASSED PAWNS - BLACK DIST.= " << b_out_pp << endl;
#endif
        if (w_out_pp != b_out_pp)
        {
            int o_score = score_out_ppawns[abs(w_out_pp - b_out_pp)]*(stage-1);

            if (w_out_pp < b_out_pp) o_score = -o_score;
            sscore +=  o_score;

#ifdef DEBUG_EVALPAWNS
            cout << "OUTSIDEPASSED PAWNS - SCORE = " << o_score << endl;
#endif
        }
    }

    // ----------------------------------------------------------------------
    // In Abhngigkeit von der Spielphase den Strukturwert skalieren:
    //
    // score = "Positionelle Score" + FAKTOR(stage) * "Strukturelle Score"
    //
    // ----------------------------------------------------------------------

    switch (stage)
    {
    case OPENING:
        pawn_info.score = pscore + ((sscore* SCALE_PWN_OP) /64);
        break;
    case EARLY_MIDGAME:
        pawn_info.score = pscore + ((sscore* SCALE_PWN_EM) /64);
        break;
    case LATE_MIDGAME:
        pawn_info.score = pscore + ((sscore* SCALE_PWN_LM) /64);
        break;
    case ENDGAME:
        pawn_info.score = pscore + ((sscore* SCALE_PWN_EG) /64);
        break;
    }


#ifdef DEBUG_EVALPAWNS
    for (int col = 0; col < 8; col++)
    {
        if (pawn_info.isHalfOpenBlack(col))
            cout << "HALF OPEN BLACK COL = " << col << endl;
        if (pawn_info.isHalfOpenWhite(col))
            cout << "HALF OPEN WHITE COL = " << col << endl;
    }

    cout << endl;
    cout << "PAWN SCORE (POS)   = " << pscore << endl;
    cout << "PAWN SCORE (STRCT) = " << sscore << endl;
    cout << "SCORE              = " << pawn_info.score << endl;
#endif




    // ----------------------------------
    //      Score im Hash speichern
    // ----------------------------------
    p_pawn_info = pB->getPTT()->insert(pawn_info,pB->get_zp1_hash());
}

// --------------------------------------------------------------
//    bishop_eval_op_emid
// --------------------------------------------------------------
int Eval::bishop_eval(const Board *pB, const int posval_bishop[])
{
    int bscore = 0;


    //    W e i s s e  L  u f e r 
    Bitboard set = pB->whiteBishops;
    while (set)
    {
        int pos = set.msb();
        set ^= BBTables::bbmask[pos];

        // statische Score
        bscore += posval_bishop[pos];

        // Abstand zum gegnerischen Knig
        int d = BBTables::dist[pos][pB->blackKingPos];
        if (d <= 3)
        {
            bscore += (4-d)*BISHOP_KING_ATTACK;
        }

#ifdef DEBUG_EVAL
        cout << "WHITE BISHOP STATIC: " << pos;
        cout << "  [ " <<  posval_bishop_op_emid[pos] << " ]" << endl;
#endif

        // Lufer-Falle?
        if (pos == 8 && (BBTables::bbmask[17] & pB->blackPawns)  ||
            pos ==15 && (BBTables::bbmask[22] & pB->blackPawns) )
        {
            bscore -= TRAPPED_BISHOP_PENALTY;

#ifdef DEBUG_EVAL
            cout << "TRAPPED WHITE BISHOP PENALTY FOR " << pos << endl;
#endif
        }

        // Beweglichkeit
        Bitboard all45C(pB->whitePieces45C | pB->blackPieces45C);
        int mobh8a1 = BBTables::diagH8A1_mobility[ pos ]
                                                 [ all45C.GET_DIAG_H8A1( rot45C_bitIndex[pos]) ];

        Bitboard all45AC(pB->whitePieces45AC | pB->blackPieces45AC);
        int mobh1a8 = BBTables::diagH1A8_mobility[ pos ]
                                                 [ all45AC.GET_DIAG_H8A1( rot45AC_bitIndex[pos]) ];

        bscore += (mobh8a1 + mobh1a8);


#ifdef DEBUG_EVAL
        cout << "WHITE BISHOP-MOBILITY: H1A8 = " << mobh1a8 << 
                " -- H8A1 = " << mobh8a1 << endl;
#endif


    }

    // Luferpaar
    if (pB->stats.nWhitePieces[BISHOP] >= 2)
    {
        bscore += BISHOPPAIR_BONUS;
    }

    //    S c h w a r z e  L  u f e r 
    set = pB->blackBishops;
    while (set)
    {
        int pos = set.lsb();
        set ^= BBTables::bbmask[pos];

        // statische Score
        bscore -= posval_bishop[mirror(pos)];

        // Abstand zum gegnerischen Knig
        int d = BBTables::dist[pos][pB->whiteKingPos];
        if (d <= 3)
        {
            bscore -= (4-d)*BISHOP_KING_ATTACK;
        }


#ifdef DEBUG_EVAL
        cout << "BLACK BISHOP STATIC: " << pos;
        cout << "  [ " <<  -posval_bishop_op_emid[mirror(pos)] << " ]" << endl;
#endif

        // Lufer-Falle?
        if (pos == 48 && (BBTables::bbmask[41] & pB->whitePawns)  ||
            pos == 55 && (BBTables::bbmask[46] & pB->whitePawns) )
        {
            bscore += TRAPPED_BISHOP_PENALTY;

#ifdef DEBUG_EVAL
            cout << "TRAPPED BLACK BISHOP PENALTY FOR " << pos << endl;
#endif
        }

        // Beweglichkeit
        Bitboard all45C(pB->whitePieces45C | pB->blackPieces45C);
        int mobh8a1 = BBTables::diagH8A1_mobility[ pos ]
                                                 [ all45C.GET_DIAG_H8A1( rot45C_bitIndex[pos]) ];

        Bitboard all45AC(pB->whitePieces45AC | pB->blackPieces45AC);
        int mobh1a8 = BBTables::diagH1A8_mobility[pos]
                                                 [ all45AC.GET_DIAG_H8A1( rot45AC_bitIndex[pos]) ];

        bscore -= (mobh8a1 + mobh1a8);


#ifdef DEBUG_EVAL
        cout << "BLACK BISHOP-MOBILITY: H1A8 = " << mobh1a8 << 
                " -- H8A1 = " << mobh8a1 << endl;
#endif

    }

    // Luferpaar
    if (pB->stats.nBlackPieces[BISHOP] >= 2)
    {
        bscore -= BISHOPPAIR_BONUS;
    }


    return bscore;
}

// --------------------------------------------------------------
//  Springer Evaluierung
// --------------------------------------------------------------
int Eval::knight_eval(const Board *pB, const int posval_knight[], 
                      const PTTe* const p_pawn_info, int stage)
{
    int kscore = 0;

    Bitboard set = pB->whiteKnights;
    while (set)
    {
        int pos = set.msb();
        set ^= BBTables::bbmask[pos];

        kscore += posval_knight[pos];

        // Abstand zum gegnerischen Knig
        int d = BBTables::dist[pos][pB->blackKingPos];
        if (d <= 3)
        {
            kscore += (4-d)*KNIGHT_KING_ATTACK;
        }

        // Blocker?
        if (pos>=16 && stage && (BBTables::bbmask[pos-8]&pB->blackPawns) &&
            ( p_pawn_info->getBPassed() & 1 << COL(pos)))
        {
            Bitboard temp(pB->blackPawns & BBTables::filemask[COL(pos)]);
            if (temp)
            {
                if (pos-8 == temp.msb())
                {
                    kscore += score_block_passer[ROW(pos-8)]*stage/4;
#ifdef DEBUG_EVAL
                    cout << "WHITE KNIGHT BLOCKS PASSER - SCORE = " << 
                            score_block_passer[ROW(pos-8)]*stage/4 << endl;
#endif
                }
            }
        }

#ifdef DEBUG_EVAL
        cout << "WHITE KNIGHT STATIC: " << pos;
        cout << "  [ " << posval_knight[pos] << " ]" << endl;
#endif
    }

    set = pB->blackKnights;
    while (set)
    {
        int pos = set.lsb();
        set ^= BBTables::bbmask[pos];

        kscore -= posval_knight[mirror(pos)];

        // Abstand zum gegnerischen Knig
        int d = BBTables::dist[pos][pB->whiteKingPos];
        if (d <= 3)
        {
            kscore -= (4-d)*KNIGHT_KING_ATTACK;
        }

        // Blocker?
        if (pos<48 && stage && (BBTables::bbmask[pos+8]&pB->whitePawns) &&
            ( p_pawn_info->getWPassed() & 1 << COL(pos)))
        {
            Bitboard temp(pB->whitePawns & BBTables::filemask[COL(pos)]);
            if (temp)
            {
                if (pos+8 == temp.lsb())
                {
                    kscore -= score_block_passer[7-ROW(pos+8)]*stage/4;
#ifdef DEBUG_EVAL
                    cout << "BLACK KNIGHT BLOCKS PASSER - SCORE = " << 
                            -score_block_passer[7-ROW(pos+8)]*stage/4 << endl;
#endif
                }
            }
        }

#ifdef DEBUG_EVAL
        cout << "BLACK KNIGHT STATIC: " << pos;
        cout << "  [ " << -posval_knight[mirror(pos)] << " ]" << endl;
#endif
    }

    return kscore;
}

// --------------------------------------------------------------
//  Turm Evaluierung
// --------------------------------------------------------------
int Eval::rook_eval(const Board *pB, const int posval_rook[], 
                    const PTTe* const p_pawn_info)
{
    int rscore = 0;

    // -----------------------------
    //  weisse Trme
    // -----------------------------
    int rooks_on_7 = 0;
    Bitboard set = pB->whiteRooks;
    while (set)
    {
        int pos = set.msb();
        set ^= BBTables::bbmask[pos];

        // statische Evaluierung
        rscore += posval_rook[pos];

        // Abstand zum gegnerischen Knig
        int d = BBTables::dist[pos][pB->blackKingPos];
        if (d <= 3)
        {
            rscore += (4-d)*ROOK_KING_ATTACK;
        }


        // Turm auf 7. Reihe?
        if (ROW(pos) == 1)
            rooks_on_7++;


#ifdef DEBUG_EVAL
        cout << "WHITE ROOK STATIC: " << pos;
        cout << "  [ " << posval_rook[pos] << " ]" << endl;
#endif

        // (Halb-)offene Linien
        int c = COL(pos);

        if (p_pawn_info->isHalfOpenWhite(c))
        {
            if (p_pawn_info->isHalfOpenBlack(c))
            {
                // Offene Linie
                rscore += ROOK_ON_OPEN_FILE;


#ifdef DEBUG_EVAL
                cout << "WHITE ROOK ON OPEN FILE: " << c << endl;
#endif
            }
            else
            {
                // Halboffene Linie
                rscore += ROOK_ON_HALFOPEN_FILE;


#ifdef DEBUG_EVAL
                cout << "WHITE ROOK ON HALFOPEN FILE: " << c << endl;
#endif
            }
        }

        // 2 Trme in 7. Reihe?
        if (rooks_on_7 == 2)
            rscore += ROOKS_ON_7;


        // Beweglichkeit
        Bitboard all(pB->whitePieces|pB->blackPieces);
        int mobx = BBTables::rank_mobility[pos][ all.GET_RANK(pos) ];

        Bitboard all90C(pB->whitePieces90C|pB->blackPieces90C);
        int moby = BBTables::file_mobility[pos][ all90C.GET_RANK(rot90C_bitIndex[pos]) ];

        rscore += (mobx+moby); 

        // Eingesperrter Turm?
        if (mobx+moby <= 1)
        {
            if (mobx+moby == 0)
                rscore -= FENCED_ROOK_PENALTY;
            else
                rscore -= FENCED_ROOK_PENALTY>>1;
        }


#ifdef DEBUG_EVAL
        cout << "WHITE ROOK-MOBILITY: RX = " << mobx << " -- RY = " << moby << endl;
#endif

    }

    // -----------------------------
    //  schwarze Trme
    // -----------------------------
    rooks_on_7 = 0;
    set = pB->blackRooks;
    while (set)
    {
        int pos = set.lsb();
        set ^= BBTables::bbmask[pos];

        rscore -= posval_rook[mirror(pos)];

        // Abstand zum gegnerischen Knig
        int d = BBTables::dist[pos][pB->whiteKingPos];
        if (d <= 3)
        {
            rscore -= (4-d)*ROOK_KING_ATTACK;
        }


        // Turm auf 7. Reihe?
        if (ROW(pos) == 6)
            rooks_on_7++;


#ifdef DEBUG_EVAL
        cout << "BLACK ROOK STATIC: " << pos;
        cout << "  [ " << -posval_rook[mirror(pos)] << " ]" << endl;
#endif
        // (Halb-)offene Linien
        int c = COL(pos);

        if (p_pawn_info->isHalfOpenBlack(c))
        {
            if (p_pawn_info->isHalfOpenWhite(c))
            {
                // Offene Linie
                rscore -= ROOK_ON_OPEN_FILE;


#ifdef DEBUG_EVAL
                cout << "BLACK ROOK ON OPEN FILE: " << c << endl;
#endif
            }
            else
            {
                // Halboffene Linie
                rscore -= ROOK_ON_HALFOPEN_FILE;


#ifdef DEBUG_EVAL
                cout << "BLACK ROOK ON HALFOPEN FILE: " << c << endl;
#endif
            }
        }

        // 2 Trme in 7. Reihe?
        if (rooks_on_7 == 2)
            rscore -= ROOKS_ON_7;


        // Beweglichkeit
        Bitboard all(pB->whitePieces|pB->blackPieces);
        int mobx = BBTables::rank_mobility[pos][ all.GET_RANK(pos) ];

        Bitboard all90C(pB->whitePieces90C|pB->blackPieces90C);
        int moby = BBTables::file_mobility[pos][ all90C.GET_RANK(rot90C_bitIndex[pos]) ];

        rscore -= (mobx+moby); 


        // Eingesperrter Turm?
        if (mobx+moby <= 1)
        {
            if (mobx+moby == 0)
                rscore += FENCED_ROOK_PENALTY;
            else
                rscore += FENCED_ROOK_PENALTY>>1;
        }




#ifdef DEBUG_EVAL
        cout << "BLACK ROOK-MOBILITY: RX = " << mobx << " -- RY = " << moby << endl;
#endif
    }


    return rscore;
}


// --------------------------------------------------------------
//  Damen Evaluierung
// --------------------------------------------------------------
int Eval::queen_eval(const Board *pB)
{
    int qscore = 0;
    // w e i s s e  D a m e n
    Bitboard set = pB->whiteQueens;
    while (set)
    {
        int pos = set.msb();
        set ^= BBTables::bbmask[pos];

        // Abstand zum schwarzen Knig
        int d = BBTables::dist[pos][pB->blackKingPos];
        if (d <= 3)
        {
            qscore += (4-d)*(4-d)*QUEEN_KING_ATTACK;
        }
    }

    // Dame + Turm auf 7.Reihe?
    if ((BBTables::rankmask[1] & pB->whiteQueens) && 
        (BBTables::rankmask[1] & pB->whiteRooks) )
            qscore += QUEEN_ROOK_ON_7;

    // s c h w a r z e  D a m e n
    set = pB->blackQueens;
    while (set)
    {
        int pos = set.lsb();
        set ^= BBTables::bbmask[pos];

        // Abstand zum schwarzen Knig
        int d = BBTables::dist[pos][pB->whiteKingPos];
        if (d <= 3)
        {
            qscore -= (4-d)*(4-d)*QUEEN_KING_ATTACK;
        }
    }

    // Dame + Turm auf 7.Reihe?
    if ((BBTables::rankmask[6] & pB->blackQueens) && 
        (BBTables::rankmask[6] & pB->blackRooks) )
            qscore -= QUEEN_ROOK_ON_7;


#ifdef DEBUG_EVAL
    cout << "QUEEN EVAL: " << qscore << endl;
#endif

    return qscore;
}



// --------------------------------------------------------------
//  Knigs Evaluierung
// --------------------------------------------------------------
int Eval::king_eval(const Board *pB, const int posval_king[])
{
    int kscore = 0;

    // w e i s s e r  K  n i g

    // statische Evaluierung
    kscore += posval_king[pB->whiteKingPos];


#ifdef DEBUG_EVAL
    cout << "WHITE KING STATIC: " << pB->whiteKingPos;
    cout << "  [ " << posval_king[pB->whiteKingPos] << " ]" << endl;
#endif



    // s c h w a r z e r  K  n i g

    // statische Evaluierung
    kscore -= posval_king[mirror(pB->blackKingPos)];


#ifdef DEBUG_EVAL
    cout << "BLACK KING STATIC: " << pB->blackKingPos;
    cout << "  [ " << -posval_king[mirror(pB->blackKingPos)] << " ]" << endl;
#endif


    return kscore;
}

// --------------------------------------------------------------
// Sicherheit der Knigsstellung
// --------------------------------------------------------------
int Eval::king_safety_score(const Board *pB, const PTTe * const p_pawn_info)
{
    int ks_score = 0;
    int c;

    // Falls Knige im feindlichen Lager sind: Aufgeben ;)
    if (pB->whiteKingPos < 16 || pB->blackKingPos > 47)
        return 0;

    // -----------------------------------
    //      w e i s s e r  K  n i g
    // -----------------------------------

    // offene Linien?
    int pos = pB->whiteKingPos;
    c = COL(pos);
    if ((pB->stats.nBlackPieces[ROOK] || pB->stats.nBlackPieces[QUEEN] )
        && p_pawn_info->isHalfOpenWhite(c))
    {
        if (p_pawn_info->isHalfOpenBlack(c))
        {
            ks_score -= KING_ON_OPEN_FILE;
        }
        else
        {
            ks_score -= KING_ON_HALFOPEN_FILE;
        }
#ifdef DEBUG_EVAL
        cout << "WHITE KING ON OPEN FILE: " << c << endl;
#endif
    }

    // Rochademglichkeiten?
    // Strafe, falls Knig Rochade rechte verloren hat
    if (!pB->whiteKingCastled)
    {
        if ( ! ( pB->castle & (CASTLE_WHITE_KINGSIDE | CASTLE_WHITE_QUEENSIDE)) )
        {
            ks_score -= NO_CASTLE_PENALTY;


#ifdef DEBUG_EVAL
            cout << "NO CASTLE PENALTY WHITE" << endl;
#endif
        }
    }

    // Strke Bauern Schutzwall?
    int safety = 0;
    if (BBTables::bbmask[pos - 8] & pB->whitePawns)
        safety += 4;
    else
        if (BBTables::bbmask[pos - 16] & pB->whitePawns)
            safety += 2;
    if (c > 1)
        if (BBTables::bbmask[pos - 9] & pB->whitePawns)
            safety += 2;
        else if (BBTables::bbmask[pos - 17] & pB->whitePawns)
            safety++;
    if (c < 7)
        if (BBTables::bbmask[pos - 7] & pB->whitePawns)
            safety +=2;
        else if (BBTables::bbmask[pos - 15] & pB->whitePawns)
            safety++;

#ifdef DEBUG_EVAL
    cout << "SAFETY PAWNWALL WHITE KING: " << safety << endl;
#endif
    if (safety <= 3)
    {
        ks_score -= CRITICAL_PAWN_WALL;
    }
    else if (safety <= 5)
    {
        ks_score -= BAD_PAWN_WALL;
    }
    else if (safety <= 6)
    {
        ks_score -= MEDIUM_PAWN_WALL;
    }
    else if (safety <= 7)
    {
        ks_score -= GOOD_PAWN_WALL;
    }
        
    // -------------------------------------
    //      s c h w a r z e r  K  n i g
    // -------------------------------------
    pos = pB->blackKingPos;
    c = COL(pos);
    if ((pB->stats.nWhitePieces[ROOK] || pB->stats.nWhitePieces[QUEEN])
         && p_pawn_info->isHalfOpenBlack(c))
    {
        if (p_pawn_info->isHalfOpenWhite(c))
        {
            ks_score += KING_ON_OPEN_FILE;
        }
        else
        {
            ks_score += KING_ON_HALFOPEN_FILE;
        }

#ifdef DEBUG_EVAL
        cout << "BLACK KING ON OPEN FILE: " << c << endl;
#endif
    }


    // Rochademglichkeiten?
    // Strafe, falls Knig Rochade rechte verloren hat
    if (!pB->blackKingCastled)
    {
        if ( ! ( pB->castle & (CASTLE_BLACK_KINGSIDE | CASTLE_BLACK_QUEENSIDE)) )
        {
            ks_score += NO_CASTLE_PENALTY;


#ifdef DEBUG_EVAL
            cout << "NO CASTLE PENALTY BLACK" << endl;
#endif
        }
    }


    // Strke Bauern Schutzwall?
    safety = 0;
    if (BBTables::bbmask[pos + 8] & pB->blackPawns)
        safety += 4;
    else
        if (BBTables::bbmask[pos + 16] & pB->blackPawns)
            safety += 2;
    if (c > 1)
        if (BBTables::bbmask[pos + 7] & pB->blackPawns)
            safety += 2;
        else if (BBTables::bbmask[pos + 15] & pB->blackPawns)
            safety++;
    if (c < 7)
        if (BBTables::bbmask[pos + 9] & pB->blackPawns)
            safety +=2;
        else if (BBTables::bbmask[pos + 17] & pB->blackPawns)
            safety++;

#ifdef DEBUG_EVAL
    cout << "SAFETY PAWNWALL BLACK KING: " << safety << endl;
#endif
    if (safety <= 3)
    {
        ks_score += CRITICAL_PAWN_WALL;
    }
    else if (safety <= 5)
    {
        ks_score += BAD_PAWN_WALL;
    }
    else if (safety <= 6)
    {
        ks_score += MEDIUM_PAWN_WALL;
    }
    else if (safety <= 7)
    {
        ks_score += GOOD_PAWN_WALL;
    }


#ifdef DEBUG_EVAL
    cout << "TOTAL KING SAFETY SCORE = " << ks_score << endl;
#endif
    return ks_score;
}



// --------------------------------------------------------------
//              e v a l _ o p
// --------------------------------------------------------------
//
//  Evaluierungsfunktion fr die Erffnung
// --------------------------------------------------------------
int Eval::eval_op(const Board* pB, int alpha, int beta)
{
    int score = 0;

    // --------------------------
    // Materialscore berechnen
    // --------------------------
    score = pB->stats.whiteMatScore;
#ifdef DEBUG_EVAL
    cout << "MATERIAL-SCORE = " << score << endl;
#endif

    // Materialverteilung?
    score += bad_trades(pB);

#ifdef DEBUG_EVAL
    cout << "TRADES = " << bad_trades(pB) << endl;
#endif
    
    // --------------------------
    //  Bauern Score
    // --------------------------

    PTTe* p_pawn_info;
    if (!(p_pawn_info = pB->getPTT()->lookup(pB->get_zp2_hash(), OPENING, pB->get_zp1_hash())))
        pawn_eval(pB,p_pawn_info,OPENING,posval_pawn_op);
    score += p_pawn_info->score;

    // --------------------------
    //  Lufer Score
    // --------------------------
    score += bishop_eval(pB, posval_bishop_op_emid);

    // --------------------------
    //  Lazy Exit?
    // --------------------------
    int lazy_score = (pB->sideToMove() == WHITE) ? score : -score;
    if (lazy_score + LAZY_EXIT_OFFSET <= alpha  ||
        lazy_score - LAZY_EXIT_OFFSET >= beta)
    {
        nLazyExits++;

#ifndef DEBUG_LAZYEVAL
        return lazy_score;
#endif
    }


    // --------------------------
    //  Springer Score
    // --------------------------
    score += knight_eval(pB,posval_knight_op_emid, p_pawn_info, OPENING);

    // --------------------------
    //  Turm Score
    // --------------------------
    score += rook_eval(pB,posval_rook_op, p_pawn_info);

    // --------------------------
    //  Knigs Score
    // --------------------------
    score += king_eval(pB,posval_king_op);

    score += king_safety_score(pB, p_pawn_info);

    // --------------------------
    // Damen-Score
    // --------------------------
    score += queen_eval(pB);


    // --------------------------
    //  Lazy Evaluation debuggen
    // --------------------------
#ifdef DEBUG_LAZYEVAL
    if (lazy_score + LAZY_EXIT_OFFSET <= alpha  ||
        lazy_score - LAZY_EXIT_OFFSET >= beta)
    {
        // Lazy Eval htte stattgefunden

        // Wre es gerechtfertigt?
        int sc = (pB->sideToMove() == WHITE) ? score : -score;
        if (sc > alpha && sc < beta)
        {
            // Fehler
            nWrongLazyExits++;
        }
    }
#endif


    if (pB->sideToMove() == WHITE)
        return score;
    else 
        return -score;
}

// --------------------------------------------------------------
//              e v a l _ e m i d
// --------------------------------------------------------------
//
//  Evaluierungsfunktion fr frhes Mittelspiel
// --------------------------------------------------------------
int Eval::eval_emid(const Board* pB, int alpha, int beta)
{
    int score = 0;

    // --------------------------
    // Materialscore berechnen
    // --------------------------
    score = pB->stats.whiteMatScore;

#ifdef DEBUG_EVAL
    cout << "MATERIAL-SCORE = " << score << endl;
#endif

    // Materialverteilung?
    score += bad_trades(pB);

#ifdef DEBUG_EVAL
    cout << "TRADES = " << bad_trades(pB) << endl;
#endif

    // --------------------------
    //  Bauern Score
    // --------------------------

    PTTe* p_pawn_info;
    if (!(p_pawn_info = pB->getPTT()->lookup(pB->get_zp2_hash(), EARLY_MIDGAME, pB->get_zp1_hash())))
        pawn_eval(pB,p_pawn_info,EARLY_MIDGAME,posval_pawn_emid);
    score += p_pawn_info->score;

    // --------------------------
    //  Lufer Score
    // --------------------------
    score += bishop_eval(pB, posval_bishop_op_emid);

    // --------------------------
    //  Lazy Exit?
    // --------------------------
    int lazy_score = (pB->sideToMove() == WHITE) ? score : -score;
    if (lazy_score + LAZY_EXIT_OFFSET <= alpha  ||
        lazy_score - LAZY_EXIT_OFFSET >= beta)
    {
        nLazyExits++;

#ifndef DEBUG_LAZYEVAL
        return lazy_score;
#endif
    }

    // --------------------------
    //  Springer Score
    // --------------------------
    score += knight_eval(pB,posval_knight_op_emid, p_pawn_info, EARLY_MIDGAME);

    // --------------------------
    //  Turm Score
    // --------------------------
    score += rook_eval(pB,posval_rook_emid, p_pawn_info);

    // --------------------------
    //  Knigs Score
    // --------------------------
    score += king_eval(pB,posval_king_emid);
    score += king_safety_score(pB, p_pawn_info);

    // --------------------------
    // Damen-Score
    // --------------------------
    score += queen_eval(pB);

    // --------------------------
    //  Lazy Evaluation debuggen
    // --------------------------
#ifdef DEBUG_LAZYEVAL
    if (lazy_score + LAZY_EXIT_OFFSET <= alpha  ||
        lazy_score - LAZY_EXIT_OFFSET >= beta)
    {
        // Lazy Eval htte stattgefunden

        // Wre es gerechtfertigt?
        int sc = (pB->sideToMove() == WHITE) ? score : -score;
        if (sc > alpha && sc < beta)
        {
            // Fehler
            nWrongLazyExits++;
        }
    }
#endif


    if (pB->sideToMove() == WHITE)
        return score;
    else 
        return -score;
}

// --------------------------------------------------------------
//              e v a l _ l m i d 
// --------------------------------------------------------------
//
//  Evaluierungsfunktion fr sptes Mittelspiel
// --------------------------------------------------------------
int Eval::eval_lmid(const Board* pB, int alpha, int beta)
{
    int score = 0;

    // --------------------------
    // Materialscore berechnen
    // --------------------------
    score = pB->stats.whiteMatScore;

#ifdef DEBUG_EVAL
    cout << "MATERIAL-SCORE = " << score << endl;
#endif

    // Materialverteilung?
    score += bad_trades(pB);

#ifdef DEBUG_EVAL
    cout << "TRADES = " << bad_trades(pB) << endl;
#endif


    // --------------------------
    //  Bauern Score
    // --------------------------

    PTTe* p_pawn_info;
    if (!(p_pawn_info = pB->getPTT()->lookup(pB->get_zp2_hash(), LATE_MIDGAME, pB->get_zp1_hash())))
        pawn_eval(pB,p_pawn_info,LATE_MIDGAME,posval_pawn_lmid);
    score += p_pawn_info->score;

    // --------------------------
    //  Lufer Score
    // --------------------------
    score += bishop_eval(pB, posval_bishop_lmid_end);


    // --------------------------
    //  Lazy Exit?
    // --------------------------
    int lazy_score = (pB->sideToMove() == WHITE) ? score : -score;
    if (lazy_score + LAZY_EXIT_OFFSET <= alpha  ||
        lazy_score - LAZY_EXIT_OFFSET >= beta)
    {
        nLazyExits++;

#ifndef DEBUG_LAZYEVAL
        return lazy_score;
#endif
    }

    // --------------------------
    //  Springer Score
    // --------------------------
    score += knight_eval(pB,posval_knight_lmid_end, p_pawn_info, LATE_MIDGAME);

    // --------------------------
    //  Turm Score
    // --------------------------
    score += rook_eval(pB,posval_rook_lmid, p_pawn_info);

    // --------------------------
    //  Knigs Score
    // --------------------------
    score += king_eval(pB,posval_king_lmid);
    score += king_safety_score(pB, p_pawn_info);

    // --------------------------
    // Damen-Score
    // --------------------------
    score += queen_eval(pB);

    // --------------------------
    //  Lazy Evaluation debuggen
    // --------------------------
#ifdef DEBUG_LAZYEVAL
    if (lazy_score + LAZY_EXIT_OFFSET <= alpha  ||
        lazy_score - LAZY_EXIT_OFFSET >= beta)
    {
        // Lazy Eval htte stattgefunden

        // Wre es gerechtfertigt?
        int sc = (pB->sideToMove() == WHITE) ? score : -score;
        if (sc > alpha && sc < beta)
        {
            // Fehler
            nWrongLazyExits++;
        }
    }
#endif


    if (pB->sideToMove() == WHITE)
        return score;
    else 
        return -score;
}

// --------------------------------------------------------------
//              e v a l _ e n d
// --------------------------------------------------------------
//
//  Evaluierungsfunktion fr das Endspiel
// --------------------------------------------------------------
int Eval::eval_end(const Board* pB, int alpha, int beta)
{
    RecogTable::Recognizer* p_rec;
    int low_b = -INFINITY;
    int upp_b = +INFINITY;
    if (pB->stats.nWhitePiecesTotal <= 4 && pB->stats.nBlackPiecesTotal <= 4 && 
        (p_rec=RecogTable::getHandle()->lookup(pB->stats.matWhite, pB->stats.matBlack))
       )
    {
        RecogTable::RecogResult res;

        if ( (res=p_rec->execute(pB)) ) // != FAILE
        {
            if (res == RecogTable::EXACT)
            {
                return p_rec->getScore();
            }
            else if (res == RecogTable::LBOUND)
            {
                low_b = p_rec->getScore();
                if (low_b >= beta)
                    return low_b;
            }
            else if (res == RecogTable::UBOUND)
            {
                upp_b = p_rec->getScore();
                if (upp_b <= alpha)
                    return upp_b;
            }

        }
    } 

    int score = 0;

    // --------------------------
    // Materialscore berechnen
    // --------------------------
    score = pB->stats.whiteMatScore;

    // Spezielle Materialverteilungen

    // R vs B  oder  R vs N
    int mat = pB->stats.matWhite|pB->stats.matBlack;
    if (mat == 13 || mat == 11)
    {
        if ( (pB->stats.matWhite&30) == 8  && pB->stats.nWhitePieces[ROOK] == 1 &&
             (((pB->stats.matBlack&30) == 4  && pB->stats.nBlackPieces[BISHOP] == 1) ||
              (pB->stats.matBlack&30) == 2  && pB->stats.nBlackPieces[KNIGHT] == 1))
        {
            score += ROOK_VS_MINOR;
        }
        else if ( (pB->stats.matBlack&30) == 8  && pB->stats.nBlackPieces[ROOK] == 1 &&
             (((pB->stats.matWhite&30) == 4  && pB->stats.nWhitePieces[BISHOP] == 1) ||
              (pB->stats.matWhite&30) == 2  && pB->stats.nWhitePieces[KNIGHT] == 1))
        {
            score -= ROOK_VS_MINOR;
        }
    }


#ifdef DEBUG_EVAL
    cout << "MATERIAL-SCORE = " << score << endl;
#endif

    // --------------------------
    //  Bauern Score
    // --------------------------

    PTTe* p_pawn_info;

    if (!(p_pawn_info = pB->getPTT()->lookup(pB->get_zp2_hash(), ENDGAME, pB->get_zp1_hash())))
    {
        pawn_eval(pB,p_pawn_info,ENDGAME,posval_pawn_end);
    }
    score += p_pawn_info->score;

    score += uncatchable_pawn_eval(pB, *p_pawn_info);


    // --------------------------
    //  Lufer Score
    // --------------------------
    if (mat&4)
        score += bishop_eval(pB, posval_bishop_lmid_end);

    // -----------------------------
    // Stellung tendiert zum Remis?
    // -----------------------------
    bool is_drawish = drawish(pB);


    // --------------------------------------------------
    // Falls es einen grossen (Material-)Vorsprung gibt,
    // ermuntere ggfalls zum Damentausch
    // --------------------------------------------------

    if (abs(score) >= 400)
    {
        if (score >= 400)
        {
            if (pB->stats.matBlack & 16) // Dame?
                score -= QUEEN_DEFENDER;
        }
        else
        {
            if (pB->stats.matWhite & 16) // Dame?
                score += QUEEN_DEFENDER;
        }
    }


    // --------------------------
    //  Lazy Exit?
    // --------------------------
    int lazy_score = (pB->sideToMove() == WHITE) ? score : -score;
    if (lazy_score > upp_b)         lazy_score = upp_b;
    else if (lazy_score < low_b)    lazy_score = low_b;
    if (!is_drawish && (lazy_score + LAZY_EXIT_OFFSET <= alpha  ||
        lazy_score - LAZY_EXIT_OFFSET >= beta))
    {
        nLazyExits++;

#ifndef DEBUG_LAZYEVAL
        return lazy_score;
#endif
    }

    // --------------------------
    //  Springer Score
    // --------------------------
    if (mat&2)
        score += knight_eval(pB,posval_knight_lmid_end, p_pawn_info, ENDGAME);

    // --------------------------
    //  Turm Score
    // --------------------------
    if (mat&8)
        score += rook_eval(pB,posval_rook_end, p_pawn_info);

    // --------------------------
    //  Knigs Score
    // --------------------------
    score += king_eval(pB,posval_king_end);

    // --------------------------
    //  Lazy Evaluation debuggen
    // --------------------------
#ifdef DEBUG_LAZYEVAL
    if (lazy_score + LAZY_EXIT_OFFSET <= alpha  ||
        lazy_score - LAZY_EXIT_OFFSET >= beta)
    {
        // Lazy Eval htte stattgefunden

        // Wre es gerechtfertigt?
        int sc = (pB->sideToMove() == WHITE) ? score : -score;
        if (sc > alpha && sc < beta)
        {
            // Fehler
            nWrongLazyExits++;
        }
    }
#endif

    if (is_drawish)
        score /= 16;

    if (pB->sideToMove() == BLACK)
        score = -score;

    if (score < low_b) return low_b;
    if (score > upp_b) return upp_b;


    return score;
}

// --------------------------------------------------------------
//  Freibauern finden, die nicht mehr gestoppt werden knnen
// --------------------------------------------------------------
int Eval::uncatchable_pawn_eval(const Board* pB, const PTTe& pawn_info)
{
    // Prfen, ob berhaupt Freibauern vorhanden sind
    unsigned int w_passed = pawn_info.getWPassed();
    unsigned int b_passed = pawn_info.getBPassed();

    if (!w_passed && !b_passed)
        return 0;

    if (pB->stats.matBlack > 1 && pB->stats.matWhite > 1)
        return 0;

    // Wieviel Schritte zur Damenumwandlung?
    int w_to_queen = 100;
    int b_to_queen = 100;

    // ---------------------------------------------------
    // Wenn Schwarz kein Material hat, weisse Freibauern
    // untersuchen
    // ---------------------------------------------------

    if (pB->stats.matBlack <= 1)
    {
        while (w_passed)
        {
            int c = BBTables::lsb16[w_passed];
            w_passed ^= 1<<c;

            Bitboard temp(pB->whitePawns & BBTables::filemask[c]);
            if (!temp)
            {
                // Moeglichen Hash-Fehler abfangen
                return 0;
            }
            int pawn_pos = temp.lsb();
            int target   = COL(pawn_pos);

            // Alles bis zum Umwandlungsfeld frei?
            if ( 
                 (BBTables::squaresBetween[pawn_pos][target] | 
                  BBTables::bbmask[target]) & 
                 (pB->whitePieces | pB->blackPieces)
               )
                continue;

            int steps = ROW(pawn_pos);
            if (pB->sideToMove() == WHITE)
            {
                if (BBTables::dist[pB->blackKingPos][target] > steps)
                {
                    if (steps < w_to_queen)
                        w_to_queen = 2*steps - 1;
                }
            }
            else
            {
                if (BBTables::dist[pB->blackKingPos][target] > steps + 1)
                {
                    if (steps < w_to_queen)
                        w_to_queen = 2*steps;
                }
            }
        }
    }
    // ---------------------------------------------------
    // Wenn Weiss kein Material hat, schwarze Freibauern
    // untersuchen
    // ---------------------------------------------------
    if (pB->stats.matWhite <= 1)
    {
        while (b_passed)
        {
            int c = BBTables::lsb16[b_passed];
            b_passed ^= 1<<c;

            Bitboard temp(pB->blackPawns & BBTables::filemask[c]);
            if (!temp)
            {
                // Hash-Fehler?
                return 0;
            }
            int pawn_pos = temp.msb();
            int target   = COL(pawn_pos) + 56;

            // Alles bis zum Umwandlungsfeld frei?
            if ( 
                 (BBTables::squaresBetween[pawn_pos][target] | 
                  BBTables::bbmask[target]) & 
                 (pB->whitePieces | pB->blackPieces)
               )
                continue;

            int steps = 7-ROW(pawn_pos);
            if (pB->sideToMove() == BLACK)
            {
                if (BBTables::dist[pB->whiteKingPos][target] > steps)
                {
                    if (steps < b_to_queen)
                        b_to_queen = 2*steps-1;
                }
            }
            else
            {
                if (BBTables::dist[pB->whiteKingPos][target] > steps + 1)
                {
                    if (steps < b_to_queen)
                        b_to_queen = steps;
                }
            }
        }
    }


    if (w_to_queen==100 && b_to_queen==100)
        return 0;

#ifdef DEBUG_EVAL
    cout << "PAWN - RACE" << endl;
    cout << "  w_to_queen == " << w_to_queen << endl;
    cout << "  b_to_queen == " << b_to_queen << endl;
#endif

    if (b_to_queen == 100)
        return piece_val[QUEEN] - 2*piece_val[PAWN] - 5*w_to_queen;

    if (w_to_queen == 100)
        return -(piece_val[QUEEN] - 2*piece_val[PAWN] - 5*b_to_queen);

    if (w_to_queen + 2 < b_to_queen)
        return piece_val[QUEEN] - 2*piece_val[PAWN] - 5*w_to_queen;

    if (b_to_queen + 2 < w_to_queen)
        return -(piece_val[QUEEN] - 2*piece_val[PAWN] - 5*b_to_queen);

    return 0;
}


// --------------------------------------------------------------
//  Erkennen von Endspielstellungen, die wahrscheinlich Remis
//  sind
// --------------------------------------------------------------

bool Eval::drawish(const Board *pB)
{
    // ----------------------------------------------------------------
    // Turm-Bauer und falscher Lufer kann (meistens) nicht gewinnen
    // ----------------------------------------------------------------
    if (pB->stats.matWhite == 5 && pB->stats.nWhitePieces[PAWN]==1)
    {
        if (pB->stats.whiteMatScore > 0)
        {
            // Turmbauer?
            if ((pB->whitePawns & (BBTables::filemask[0]|BBTables::filemask[7])) )
            {
                // Falscher Lufer?
                bool is_a_pawn = (pB->whitePawns&BBTables::filemask[0]) != 0;
                bool is_white_bishop = (pB->whiteBishops & BBTables::whiteSquares) != 0;

                if ( ( is_a_pawn && !is_white_bishop) ||
                     (!is_a_pawn &&  is_white_bishop))
                {
                    // Drawish; den Rest muss die Suche erledigen :)
                    return true;
                }
            }
        }
    }
    if (pB->stats.matBlack == 5 && pB->stats.nBlackPieces[PAWN]==1)
    {
        if (pB->stats.whiteMatScore < 0)
        {
            // Turmbauer?
            if ((pB->blackPawns & (BBTables::filemask[0]|BBTables::filemask[7])) )
            {
                // Falscher Lufer?
                bool is_a_pawn = (pB->blackPawns&BBTables::filemask[0]) != 0;
                bool is_white_bishop = (pB->blackBishops & BBTables::whiteSquares) != 0;

                if ( ( is_a_pawn &&  is_white_bishop) ||
                     (!is_a_pawn && !is_white_bishop))
                {
                    // Drawish; den Rest muss die Suche erledigen :)
                    return true;
                }
            }
        }
    }

    // Nie, wenn Bauern vorhanden:
    if ((pB->stats.matBlack | pB->stats.matWhite) & 1)
        return false;

    // Notwendig: Material <= Val(Bishop)
    if (abs(pB->stats.whiteMatScore) <= BISHOP_MATERIAL)
    {
        // Ab hier normalerweise Remis; es gibt aber Ausnahmen:
        switch (pB->stats.nBlackPiecesTotal + pB->stats.nWhitePiecesTotal)
        {
        case 5:
            // KQ x KBB  v  KQ x KBN  v  KQ x KNN
            if (pB->stats.matBlack == 16 && pB->stats.matWhite <= 6)
                return false;
            if (pB->stats.matWhite == 16 && pB->stats.matBlack <= 6)
                return false;

            // KBB x KN
            if (pB->stats.nBlackPieces[BISHOP] == 2 && pB->stats.matWhite == 2)
                return false;
            if (pB->stats.nWhitePieces[BISHOP] == 2 && pB->stats.matBlack == 2)
                return false;
            break;

        }

        return true;
    }

    return false;
}



// --------------------------------------------------------------
//  bad_trades ... Erkennen ungnstiger Materialverteilungen
// --------------------------------------------------------------
inline int Eval::bad_trades(const Board* pB)
{
    int n_whiteminor = pB->stats.nWhitePieces[BISHOP] + pB->stats.nWhitePieces[KNIGHT];
    int n_blackminor = pB->stats.nBlackPieces[BISHOP] + pB->stats.nBlackPieces[KNIGHT];

    if (pB->stats.nWhitePieces[ROOK] != pB->stats.nBlackPieces[ROOK])
    {
        if (pB->stats.nWhitePieces[ROOK] == pB->stats.nBlackPieces[ROOK] + 1)
        {
            // Weiss mit einem Turm mehr:
            if (n_whiteminor + 1 == n_blackminor)  // Weiss eine Leichtf. weniger
                return BAD_TRADE;
            if (n_whiteminor + 2 == n_blackminor)  // Weiss zwei Leichtf. weniger
                return -BAD_TRADE;
            return 0;
        }
        if (pB->stats.nWhitePieces[ROOK] + 1 == pB->stats.nBlackPieces[ROOK])
        {
            // Schwarz mit einem Turm mehr:
            if (n_blackminor + 1 == n_whiteminor) // Schwarz eine Leichtf. weniger
                return -BAD_TRADE;
            if (n_blackminor + 2 == n_whiteminor) // Schwarz zwei Leichtf. weniger
                return +BAD_TRADE;
            return 0;
        }
    }
    else if (n_whiteminor != n_blackminor)
    {
        if (n_whiteminor == n_blackminor + 1)
        {
            // Weiss mit einer Leichtfigur mehr
            return BAD_TRADE;
        }
        if (n_whiteminor + 1 == n_blackminor)
        {
            // Schwarz mit einer Leichtfigur mehr
            return -BAD_TRADE;
        }
    }

    return 0;
}
