// -------------------------------------------------------------------------
//
//                                    R E S P
//                                    =======
//
// (peter Rosendahls Erstes Schach Programm).
//
//
//  Datei              : incmovgen.h
//                       incremental movegenerator
//
//  Anfang des Projekts: So, 8.Juli, 2001
//  Sprache            : C++
//  Verwendete Compiler: Visual C++ 6.0 SP5 / gcc 3.01
//  Betriebssystem(e)  : Win2000 SP2 / Linux 2.4.4
//  Version            : 0.18
// -------------------------------------------------------------------------
//
// resp (peter _r_osendahls _e_rstes _s_chach_p_rogramm) 
// $Id: incmovgen.h,v 1.7 2002/06/11 19:04:44 Peter Rosendahl Exp $
// Copyright (C) 2001 Peter Rosendahl
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// Email: p_rosendahl@t-online.de
// -------------------------------------------------------------------------
#ifndef INCMOVGEN_H
#define INCMOVGEN_H

#include "move.h"
#include "board.h"

// -------------------------------------------------------------------------
//              c l a s s  I n c M o v G e n
// -------------------------------------------------------------------------
const int SKIP = -10000;

class IncMoveGen
{
private:
    Board*      pB;         // access to current position
    int         ply;        // current ply in the search
    Move        hashMove;   // hashmove (== 0, if not available)
    int         iter_depth; // iter_depth for move ordering

    int         curMove;    // current move

    MoveStack*  pMS;        // pointer to movelist

//#define IMG_STATS
#ifdef IMG_STATS
public:
    static int counter[9];
    static void dump();
#endif

    enum { 
        GET_HASH,
        GET_KILLER_I,
        GET_KILLER_II,
        GEN_CAP,
        GET_CAP,
        GEN_NONCAP,
        GET_NONCAP,
        GEN_EVASIONS,
        GET_EVASIONS
    } phase;                // current phase from next

public:
    // constructor
    IncMoveGen(Board* _pB, int _ply, const Move& _hashMove, int inCheck, int _iter_depth=0) : 
      pB(_pB), ply(_ply), hashMove(_hashMove), iter_depth(_iter_depth)
    {
        phase = (inCheck) ? GEN_EVASIONS : GET_HASH;

#ifdef IMG_STATS
        static bool init = false;

        if (!init)
        {
            for (int i=0; i < 9; i++) counter[i] = 0;
            init = true;
        }
#endif
    }

private:
    void score_evasions();
    void score_noncaps();
    void score_caps();


    void swapWithBestScoringMove(int idx);


public:
    Move next();        // get the next move
};

#endif // INCMOVGEN_H

