// -------------------------------------------------------------------------
//
//                                    R E S P
//                                    =======
//
// (peter Rosendahls Erstes Schach Programm).
//
//
//  Datei              : inifiles.h
//                       Optionen von ASCII-Dateien laden
//
//  Anfang des Projekts: Do, 22.11.01
//  Sprache            : C++
//  Verwendete Compiler: Visual C++ 6.0 SP5 / gcc 3.01
//  Betriebssystem(e)  : Win2000 SP2 / Linux 2.4.4
//  Version            : 0.19
// -------------------------------------------------------------------------
//
// resp (peter _r_osendahls _e_rstes _s_chach_p_rogramm) 
// $Id: inifiles.cpp,v 1.8 2002/06/11 19:04:44 Peter Rosendahl Exp $
// Copyright (C) 2001 Peter Rosendahl
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// Email: p_rosendahl@t-online.de
// -------------------------------------------------------------------------

#include "inifiles.h"
#include <iostream>
#include <map>
#include <string>

using namespace std;

// -------------------------------------------------------------------------
//                  c l a s s   I n i F i l e M a n a g e r
// -------------------------------------------------------------------------


// -------------------------------------------------------------------------
//  Konstruktor
// -------------------------------------------------------------------------

IniFileManager::IniFileManager(const std::string& filename, bool _talk) : 
    infilename(filename), talk(_talk)
{
    
}

// -------------------------------------------------------------------------
//  Parsen einer INI-Datei
// -------------------------------------------------------------------------

bool IniFileManager::parse()
{
    infile.open(infilename.c_str());

    if (!infile)
    {
        return false;
    }


    // Zeilenweise scannen

    int linecount = 0;
    bool no_errors = true;
    cur_groupid = DEFAULT_GROUP_ID;
    while (!infile.eof())
    {
        string ini_line;
        getline(infile, ini_line);
        linecount++;

        if (!parseLine(ini_line))
        {
            if (talk)
                cout << "WARNING: PARSE ERROR IN LINE " << linecount << endl;
            no_errors = false;
        }
    }



    infile.close();

    return no_errors;
}


bool IniFileManager::parseLine(std::string line)
{

    int pos;

    // Whitespace berlesen
    pos = line.find_first_not_of(" \t");
    line.erase(0,pos);


    // Leerzeile?
    if (line.empty())
        return true;

    // Kommentar?
    if (line[0] == ';')
        return true;

    // Falls spter inm der Zeile noch ein Kommentar auftritt, wird
    // er an dieser Stelle abgeschnitten

    pos = line.find(';');
    if (pos != -1)
    {
        line.erase(pos,line.length());
    }

    // Gruppenbezeichner?
    if (line[0] == '[')
    {
        // 1.Zeichen lschen
        line.erase(line.begin());

        // ] suchen
        pos = line.find(']');
        if (pos == -1) return false;

        string groupid(line,0,pos);

        if (groupid.empty())
            return false;

        cur_groupid = groupid;
        return true;
    }

    // Ab hier muss eine Variablendefinition vorliegen

    // Variablenname einlesen
    if (!isalpha(line[0]))
        return false;
  
    string varName;
    varName = line[0];
    line.erase(line.begin());
    while (!line.empty() && (isalnum(line[0]) || line[0] == '_') )
    {
        varName += line[0];
        line.erase(line.begin());
    }
 
    // Whitespace berlesen
    pos = line.find_first_not_of(" \t");
    if (pos == -1)
        return false;

    // ... und abschneiden
    line.erase(0,pos);
    if (line.empty())
        return false;
    
    // '=' parsen
    if (line[0] != '=')
        return false;
    line.erase(line.begin());

    // Whitespace von vorne und hinten abschneiden
    pos = line.find_first_not_of(" \t");
    line.erase(0,pos);
    pos = line.find_last_not_of(" \t");
    line.erase(pos+1,line.length());
    

    string varValue(line);
    if (varValue.empty())
        // leerer String wird wie " " behandelt
        varValue = " ";


    // EINTRAGEN: cur_groupid . varName => varValue

    string id = cur_groupid + "." + varName;

    varList[id] = varValue;

    //cout << "INSERT: [" << id << "=>" << varValue << "]" << endl;

    return true;
}

// -------------------------------------------------------------------------
//  Wert einer Variablen auslesen
// -------------------------------------------------------------------------
bool IniFileManager::getVarValue(const std::string& name, std::string& value) 
{
    map<string,string>::iterator it = varList.find(name);


    value = "";
    if (it == varList.end())
        return false;

    value = (*it).second;
    return true;
}
