// -------------------------------------------------------------------------
//
//                                    R E S P
//                                    =======
//
// (peter Rosendahls Erstes Schach Programm).
//
//
//  Datei              : inifiles.h
//                       Optionen von ASCII-Dateien laden
//
//  Anfang des Projekts: Do, 22.11.01
//  Sprache            : C++
//  Verwendete Compiler: Visual C++ 6.0 SP5 / gcc 3.01
//  Betriebssystem(e)  : Win2000 SP2 / Linux 2.4.4
//  Version            : 0.19
// -------------------------------------------------------------------------
//
// resp (peter _r_osendahls _e_rstes _s_chach_p_rogramm) 
// $Id: inifiles.h,v 1.9 2002/06/11 19:04:44 Peter Rosendahl Exp $
// Copyright (C) 2001 Peter Rosendahl
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// Email: p_rosendahl@t-online.de
// -------------------------------------------------------------------------

#ifndef INIFILES_H
#define INIFILES_H

#ifndef LINUX
#pragma warning (disable:4786) // Visual C++ Faxen
#endif


#include <string>
#include <map>
#include <fstream>


// -------------------------------------------------------------------------
//                  c l a s s   I n i F i l e M a n a g e r
// -------------------------------------------------------------------------
//
//  Ein Ini-File dient dazu, Porgrammoptionen aus einer Text-Datei zu laden.
//  Folgendes Format wird dabei verlangt:
//
//  1) Kommentare fangen mit einem Semicolon ';' an, und gehen bis zum 
//     Zeilenende
//  2) Ein Gruppe besteht aus einem Gruppenbezeichner gefolgt von beliebig
//     vielen Zeilen bis zum nchsten Gruppenbezeichner oder dem Dateiende.
//     Eine Sonderrolle hat die Gruppe "default". Sie besteht aus den
//     Zeilen vom Dateianfang bis zum ersten Gruppenbezeichner.
//  3) Gruppenbezeichnerformat:
//
//     groupid -> '[' string ']' '\n'      (Whitespace und Kommentare werden
//                                          ignoriert)
//
//  4) Innerhalb einer Gruppe stehen die Variablendefinitionen; hier werden
//     den zu einer Gruppe gehrigen Variablen Werte zugewiesen. Format:
//
//     vardef -> varid [WhiteSpace]* '=' [WhiteSpace]* value [WhiteSpace]* '\n'   
//     varid  -> Letter [Letter|Digit|'_']*
//     value  -> [~WhiteSpace] .* [~WhiteSpace]
//
//  5) BEISPIEL EINES INI-FILES
//
//    example.ini:
//    --------------------------------------------------------------------------
//    ZEILE 01:     ;exam.ini - Beispiel fr eine Ini-Datei, die vom
//    ZEILE 02:     ;           IniFileManager gelesen werden kann
//    ZEILE 03:     ;
//    ZEILE 04:     ProgrammName = Beispiel Programm; <-in Gruppe default!
//    ZEILE 05:
//    ZEILE 06:     ; Verzeichnisse 1
//    ZEILE 07:     [Programm]
//    ZEILE 08:     Verzeichnis = c:\xyz\  ; muss mit \ oder / enden
//    ZEILE 09:
//    ZEILE 10:     ; Verzeichnisse 2
//    ZEILE 11:     [DEMO]
//    ZEILE 12:     Verzeichnis = .\demo\  ; siehe oben
//    ZEILE 13:     Dauer       = 123
//    ZEILE 14:     Name        = Irgendein Demo
//    --------------------------------------------------------------------------
//
//  6) Zugriff auf die Variablen:
//     Das Clientprogramm kann auf die Variablenwerte zugreifen, indem 
//     Gruppenname + Punkt + Variablenname an die Funktion getVarValue bergeben
//     werden. Bsp:
//
//      getVarValue("default.ProgrammName",s)  -> s == "Beispiel Programm"
//      getVarValue("Programm.Verzeichnis",s)  -> s == "c:\xyz\"
//      getVarValue("DEMO.Verzeichnis",s)      -> s == ".\demo\"
//      getVarValue("DEMO.Dauer",s)            -> s == "123"
//      getVarValue("DEMO.Name",s)             -> s == "Irgendein Demo"

const char* const DEFAULT_GROUP_ID = "default"; 
class IniFileManager
{
private:
    std::map<std::string,std::string> varList;
    
    // INI-Datei
    std::ifstream infile;
    std::string infilename;

    // Fehlermeld. nach stdout?
    bool talk;

    // Name der aktuell geparseden Gruppe
    std::string cur_groupid;
    
    // Zeile parsen
    bool parseLine(std::string line);

public:
    IniFileManager(const std::string& filename, bool _talk = true);

    // Parsen eines Ini-Files (Format s.o.)
    bool parse();

    // Wert einer Variablen auslesen
    bool getVarValue(const std::string& name, std::string& value) ;
};


#endif // INIFILES_H

